\def\topofcontents{
  \null\vfill
  \centerline{\titlefont \title}
  \vskip 15pt
  \centerline{Marcel K. de Carli Silva}
  \centerline{IME--USP, S\~ao Paulo, 30 de setembro de 2001}
  \vskip 40pt \vfill
}

\def\startsection{
   \noindent{} \bigskip \bigskip
   \let\startsection=\stsec\stsec
}

@*1 O problema.
O problema das $n$ rainhas consiste em, dado um tabuleiro de
xadrez $n \times n$, encontrar uma configura\c c\~ao (ou
todas) em que temos $n$ rainhas no tabuleiro de forma que
nenhuma rainha possa atacar nenhuma das outras no pr\'oximo
lance.

Neste programa, esse problema ser\'a ``atacado'' pelo m\'etodo
de recurs\~ao e retrocesso ({\it backtracking}).

@*2 Vis\~ao geral do programa.
Nesta se\c c\~ao est\'a exposto o esqueleto do programa.
Em primeiro lugar inclu\'\i mos todos os arquivos {\it header} 
necess\'arios e as vari\'aveis globais. Depois est\~ao as
defini\c c\~oes de todas as fun\c c\~oes do programa.

Dentro de |main|,
primeiro validamos a entrada do usu\'ario (todos os
dados s\~ao passados pela linha de comando num formato que
ser\'a especificado mais tarde). Depois, resolvemos o problema
em si (verificando se o usu\'ario quer apenas uma solu\c c\~ao
ou todas as poss\'\i veis para um dado $n$). Finalmente, 
apresentamos a sa\'\i da, que pode ser na sa\'\i da padr\~ao
ou num arquivo no formato {\tt LATEX}, dependendo da linha
de comando.

@c
@<Inclus\~ao de arquivos header@>@;
@<Vari\'aveis globais@>@;
@<Defini\c c\~oes de fun\c c\~oes@>@;
int main (int argc, char *argv[])
{
  @<Declara\c c\~ao de vari\'aveis locais@>@;
  @<Processamento da linha de comando@>@;
  @<Resolu\c c\~ao do problema@>@;
  @<Sa\'\i da da solu\c c\~ao@>@;
  return 0;
}

@ Como o programa \'e bem simples, ele apenas l\^e da entrada
padr\~ao, escreve (possivelmente) a sa\'\i da para um arquivo
e aloca dinamicamente a matriz que representar\'a o tabuleiro,
vamos declarar os arquivos {\it header} \.{stdio.h} e \.{stdlib.h}.

Al\'em disso, o programa tamb\'em cronometrar\'a o tempo que
leva para resolver uma dada inst\^ancia do programa. Por isso,
ser\'a tamb\'em inclusa a biblioteca \.{time.h}.

@<Inc...@> =
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

@ Como j\'a foi observado, o tabuleiro ser\'a armazenado,
durante a resolu\c c\~ao do problema, na forma de uma matriz
representando o tabuleiro. E essa matriz precisa ser alocada
dinamicamente para ser do tamanho $n \times n$, com $n$ 
especificado pelo usu\'ario. Por uma quest\~ao de economia de 
espa\c co, usaremos uma matriz do tipo \.{char}.

@<Var...@> =
char **tabuleiro;

@ Caso o usu\'ario pe\c ca por todas as solu\c c\~oes,
queremos salv\'a-las \`a medida que as encontramos,
para depois escrever todas no arquivo de sa\'\i da.
Para tal, criaremos uma lista ligada de vetores, cada
vetor com $n$ inteiros (na verdade, $n$ \.{char}s),
onde $v_i$ conter\'a a coluna em que se encontra a
rainha na linha $i$. Cada item da lista ligada
representa uma solu\c c\~ao. Utilizaremos uma lista
encabe\c cada e tamb\'em uma vari\'avel |num_sol| que
contar\'a o n\'umero de solu\c c\~oes encontradas.
$n$ ser\'a uma vari\'avel global e |todas| define
se o usu\'ario quer todas ou uma \'unica solu\c c\~ao.

@<Var...@> =
typedef struct _sol
{
  char *v;
  struct _sol *prox;
} solucao;
solucao *ini;
int num_sol, n;
int todas;

@*2 Entrada. Todos os par\^ametros do programa ser\~ao a ele passados
atrav\'es da linha de comando. Pelo menos um argumento \'e 
necess\'ario: o $n$. Se esse argumento n\~ao for passado, escrevemos
na sa\'\i da padr\~ao a forma de uso correta do programa e 
sa\'\i mos. Sen\~ao, guardamos o $n$ e verificamos se o usu\'ario
entrou com outras op\c c\~oes v\'alidas.

@<Processamento...@> =
if (argc < 2)
{
  fprintf (stdout, "Uso: %s n [-t] [-f <arquivo>]\n", argv[0]);
  fprintf (stdout, "     n: tamanho do tabuleiro\n");
  fprintf (stdout, "     -t: caso queira todas as possibilidades\n");
  fprintf (stdout, "     -f: para guardar os resultados num arquivo\n");
  exit (-1);
}
n = atoi (argv[1]);
@<Verifica outras op\c c\~oes de entrada@>@;
@<Corre\c c\~ao de argumentos inv\'alidos@>@;

@ Antes de verificarmos as outras entradas, vamos declarar as
vari\'aveis que definem se o usu\'ario quer as respostas 
num arquivo e \'\i ndices para percorrer as entradas. 
E tamb\'em um string que conter\'a o nome do arquivo.

@<Declara...@> =
int arquivo = 0;
int i, j, k;
char *nome_arq = NULL;

@ Agora verificamos as outras entradas, isto \'e, se o usu\'ario
quer todas as solu\c c\~oes poss\'\i veis e se as quer num arquivo.

No caso de a op\c c\~ao \.{-t} ser especificada, mesmo que o 
usu\'ario n\~ao especifique um arquivo, a sa\'\i da ser\'a 
direcionada para um arquivo \.{LATEX} de nome \.{nrainhas}$n$\.{.tex},
pois geralmente as solu\c c\~oes s\~ao muitas e seria
um exagero mandar todas para a sa\'\i da padr\~ao.

Primeiro verificamos se o argumento come\c ca com um sinal de
menos e depois verificamos a op\c c\~ao. 

@<Verifica...@> =
for (i = 2; i < argc; i++)
{
  if (argv[i][0] == '-')
  {
    if (argv[i][1] == 't')
    todas = 1;
    else if (argv[i][1] == 'f')
    arquivo = 1;
    else
    fprintf (stderr, "%s: argumento desconhecido, ignorando.\n",
	     argv[i]);
  }
  else if (arquivo)
  {
    nome_arq = argv[i];
  }
  else
  {
    fprintf (stderr, "%s: argumento desconhecido, ignorando.\n",
	     argv[i]);
  }
}

@ Caso o usu\'ario passe argumentos incorretos, como por exemplo
especificar a op\c c\~ao \.{-f} e n\~ao especificar um arquivo,
procederemos da seguinte forma: se o usu\'ario tamb\'em 
especificou \.{-t}, a sa\'\i da ir\'a para um arquivo de nome
padr\~ao \.{nrainhas}$n$\.{.tex}. Sen\~ao, ir\'a para a sa\'\i da
padr\~ao;

@<Corre...@> =
if (arquivo && !nome_arq)
{
  if (todas)
  {
    nome_arq = (char*) calloc (256, sizeof (char));
    if (!nome_arq)
    {
      fprintf (stderr, "%s: falha na alocacao de memoria\n", argv[0]);
      exit (-2);
    }
    sprintf (nome_arq, "nrainhas%d.tex", n);
  }
  else
  {
    arquivo = 0;
    fprintf (stdout, "Arquivo nao especificado. Imprimindo na saida padrao.\n");}
}
if (todas && !arquivo)
{
  arquivo = 1;
    nome_arq = (char*) calloc (256, sizeof (char));
    if (!nome_arq)
    {
      fprintf (stderr, "%s: falha na alocacao de memoria\n", argv[0]);
      exit (-2);
    }
  sprintf (nome_arq, "nrainhas%d.tex", n);
}

@*2 Resolvendo o problema. Para a resolu\c c\~ao do problema,
vamos dispor da fun\c c\~ao |resolve|. Como ela funciona 
exatamente ser\'a explicado mais tarde. 

Antes de chamar a fun\c c\~ao que resolve, ativamos o 
cron\^ometro.

@<Resolu...@> =
@<Inicializa as vari\'aveis@>@;
@<Inicia o cron\^ometro@>@;
fprintf (stdout, "Procurando resposta...aguarde\n");
fflush (stdout);
resolve(0);
@<Finaliza o cron\^ometro@>@;

@ A inicializa\c c\~ao das vari\'aveis consiste em zerar 
|num_sol| e garantir que o ponteiro |prox| da cabe\c ca
da lista ligada de solu\c c\~oes seja aterrada. Tamb\'em
\'e necess\'ario alocar a mem\'oria para o tabuleiro.
Caso alguma aloca\c c\~ao falhe, sa\'\i mos imediatamente
do programa, comunicando o erro ao usu\'ario (note que
n\~ao \'e necess\'ario desalocar, j\'a que o programa
ser\'a finalizado caso alguma aloca\c c\~ao falhe).

@<Inicializa...@> =
num_sol = 0;
ini = (solucao *) calloc (1, sizeof (solucao));

  if (!ini)
  {
    fprintf (stderr, "%s: falha na alocacao da memoria.\n", argv[1]);
    exit (-3);
  }

ini->prox = NULL;

tabuleiro = (char**) calloc (n, sizeof (char*));

if (!tabuleiro)
{
  fprintf (stderr, "%s: falha na alocacao de memoria\n",
	   argv[0]);
  exit (-2);
}  

for (i = 0; i < n; i++)
     tabuleiro[i] = (char*) calloc (n, sizeof (char));

if (i != n)
{
  fprintf (stderr, "%s: falha na alocacao de memoria\n",
	   argv[0]);
  exit (-2);
}

@ Para marcar o tempo de busca de solu\c c\~oes, precisamos
primeiro declarar tr\^es vari\'aveis:

@<Declara...@> =
double elapsed;
clock_t start, end;

@ Para iniciar o cron\^ometro, basta uma simples chamada 
de fun\c c\~ao:

@<Inicia o...@> =
start = clock();

@ Ao finalizarmos o cron\^ometro, j\'a calculamos o tempo gasto:

@<Finaliza...@> =
end = clock();
elapsed = ((double) (end - start)) / CLOCKS_PER_SEC;

fprintf (stdout, "Procura terminada: Tempo de procura: %f segs\n",
	 elapsed);

@ Agora vamos finalmente \`as fun\c c\~oes que resolvem 
o problema:

@<Defini...@> =
void salva (void)
{
  @<Guarda resposta@>@;
}

int pode (int linha, int coluna)
{
  @<Pode@>@;
}

int resolve (int linha)
{
  @<Resolve uma@>@;
}


@*3 Pode. A fun\c c\~ao pode verifica se \'e poss\'\i vel colocar uma
rainha na posi\c c\~ao (|linha|,|coluna|), isto \'e, se n\~ao h\'a 
nenhuma outra rainha no tabuleiro que ameace essa coordenada.
Tudo o que fazemos \'e buscar outras rainhas na horizontal,
vertical e diagonais contendo esse quadrado.

@<Pode@> =
int i, j;

if (tabuleiro[linha][coluna]) return 0;

for (i = 0; i < n; i++)
     if (tabuleiro[i][coluna])
	 return 0;
	 
for (j = 0; j < n; j++)
if (tabuleiro[linha][j]) 
    return 0;


    if (linha <= coluna) /* procura na diagonal principal */
{
  for (i = 0, j = coluna - linha; j < n; i++, j++)
    if (tabuleiro[i][j])
      return 0;
}
    else
{
  for (j = 0, i = linha - coluna; i < n; i++, j++)
    if (tabuleiro[i][j])
      return 0;
}

if (linha <= (n - 1 - coluna)) /* procura na diagonal secundaria */
{
  for (i = 0, j = (linha + coluna); j >= 0; j--, i++)
    if (tabuleiro[i][j])
      return 0;
}
else
{
  for (j = (n - 1), i = (linha + coluna + 1 - n); i < n; i++, j--)
    if (tabuleiro[i][j])
      return 0;
}

return 1;

@*3 Resolve. A fun\c c\~ao |resolve| tenta resolver o problema a partir de
uma dada |linha|. Por isso a fun\c c\~ao |main| chama |resolve|
com o par\^ametro $0$. O que a fun\c c\~ao faz \'e colocar
uma rainha na primeira posi\c c\~ao poss\'\i vel da |linha| e chamar
a fun\c c\~ao |resolve| para a linha seguinte. Se n\~ao for
poss\'\i vel achar uma solu\c c\~ao dessa maneira (se |resolve|
para a linha seguinte retornar $0$), colocamos a rainha na posi\c c\~ao
seguinte de |linha| e novamente tentamos resolver para a linha seguinte.

Caso o usu\'ario s\'o queira uma solu\c c\~ao, a recurs\~ao acaba 
quando \'e poss\'\i vel p\^or uma dama numa
certa coluna da linha $n - 1$, a \'ultima linha do tabuleiro. 
A\'\i{}, guarda-se a solu\c c\~ao e retorna-se 1, que retrocede at\'e
a primeira chamada. Se o usu\'ario quer todas as respostas, apenas
guarda-se a resposta na lista, retira-se a rainha e continua testando.

Quando se percorreram todas as colunas de |linha|
e n\~ao se conseguiu resolver, retorna-se $0$, indicando que n\~ao
foi poss\'\i vel achar uma solu\c c\~ao.

@<Resolve u...@> =
int col, i, j;

for (col = 0; col < n; col++)
{
  if (pode (linha, col))
    {
      tabuleiro[linha][col] = 1; /* coloca rainha */

      if (linha == (n - 1))
	{
	  num_sol++;
	  salva();
	  if (todas)
	    {
	      tabuleiro[linha][col] = 0;
	      continue;
	    }
	  else
	    return 1;
	}

      if (resolve (linha + 1))
	return 1; /* retrocede o sucesso */
      else
	{
	  tabuleiro[linha][col] = 0; /* tira a rainha */
	  continue; /* tenta outras colunas */
	}
    }
}
return 0; /* esgotaram-se as possibilidades */

@*3 Salva. A fun\c c\~ao |salva| consiste em alocar mem\'oria
para um novo item da lista, alocar mem\'oria para um
vetor de $n$ \.{char}s para guardar as posi\c c\~oes
das rainhas e guard\'a-las:

@<Gua...@> =
int i, j;
solucao * novo;

novo = (solucao *) calloc (1, sizeof (solucao));

if (!novo)
{
  fprintf (stderr, "Falha na alocacao de memoria\n");
  exit (-3);
}  

novo->v = (char *) calloc (n, sizeof (char));

if (!novo->v)
{
  fprintf (stderr, "Falha na alocacao de memoria\n");
  exit (-2);
}  

for (i = 0; i < n; i++)
{
  for (j = 0; j < n; j++)
    if (tabuleiro[i][j])
      break;
  novo->v[i] = j; /* salva a coluna */
}

novo->prox = ini->prox;
ini->prox = novo; /* insere |novo| na lista */

@*2 Resultados. Agora que o problema j\'a foi resolvido,
vamos mostrar os resultados ao usu\'ario. Se n\~ao foram
encontradas solu\c c\~oes, n\~ao nos preocuparemos em
escrev\^e-las a um arquivo, apenas informamos o usu\'ario
pela sa\'\i da padr\~ao:

@<Sa...@> =
if (!num_sol)
{
  fprintf (stdout, "Nao foram encontrada solucoes.\n\n");
  exit (0);
}
if (arquivo)
{
  @<Configura o arquivo \.{LATEX}@>@;
  if (todas)
    {
      @<Mensagem inicial todas@>@;
      @<Imprime todas@>@;
    }
  else
    {
      @<Mensagem inicial uma@>@;
      @<Imprime uma@>@;
    }
  @<Fecha arquivo@>@;
}
else
{
  @<Imprime na tela@>@;

}



@ Antes vamos definir mais uma vari\'avel global, que
ser\'a o nosso {\it file descriptor}:

@<Vari...@> =
FILE *fp;

@ Tamb\'em definimos mais uma vari\'avel local a |main|,
  que percorre as solu\c c\~oes:

@<Decl...@> =
solucao * sol;

@*3 Tela. Imprimir na tela \'e muito simples. Lembramos aqui
que s\'o imprimimos o resultado na tela se s\'o houver
uma \'unica solu\c c\~ao:

@<Imprime na tel...@> =
sol = ini->prox;
fprintf (stdout, "\nSolucao:\n");
for (i = 0; i < n; i++)
{
  for (j = 0; j < n; j++)
    {
      if (j == sol->v[i])
	fprintf (stdout, "# ");
      else
	fprintf (stdout, "0 ");
    }
  fprintf (stdout, "\n"); /* acabou coluna, pular linha */
}
fprintf (stdout, "\n");

@*3 Arquivo. Primeiro devemos abrir um arquivo (nota: 
para simplifica\c c\~ao do c\'odigo, o programa n\~ao
verifica se j\'a h\'a alguma arquivo com o mesmo nome,
ele simplesmente cria um se n\~ao houver ou sobrescreve
se j\'a houver) e escrever uns comandos b\'asicos:

@<Config...@> =
fp = fopen (nome_arq, "w");
if (!fp)
{
  fprintf (stderr, "Nao foi possivel abrir o arquivo %s.\n", nome_arq);
  exit (-4);
}

fprintf (fp, "\\documentclass[10pt,a4paper]{article}\n");
fprintf (fp, "\\author{Marcel K. de Carli Silva}\n");
fprintf (fp, "\\title{Solu\\c c\\~ao para o problema das $n$ rainhas com $n = %d$}\n", n);
fprintf (fp, "\\begin{document}\n\\maketitle\n");

@ Caso o usu\'ario s\'o quer uma solu\c c\~ao, mostramos uma 
mensagem gen\'erica:

@<Mensagem inicial u...@> =
fprintf (fp, "Abaixo est\\'a uma solu\\c c\\~ao para o problema das $n$ ");
fprintf (fp, "rainhas com $n = %d$:\n\n", n);

@ J\'a se o usu\'ario quer todas, imprimir antes o n\'umero de solu\c c\~oes
encontradas:

@<Mensagem inicial t...@> =
fprintf (fp, "H\\'a $%d$ solu\\c c\\~oes poss\\'\\i veis para o ", num_sol);
fprintf (fp, "problema das $n$ rainhas com $n = %d$. ", n);
fprintf (fp, "Todas est\\~ao listadas abaixo:\n\n");

@ Para imprimir uma, criamos uma tabela e imprimimos como na tela:

@<Imprime u...@> =
fprintf (fp, "\\begin{center}\n");
fprintf (fp, "\\begin{tabular}{|");
for (i = 0; i < n; i++)
{
     fprintf (fp, "c|"); 
} /* estabelece quantidade de colunas */
fprintf (fp, "}\n");
fprintf (fp, "\\hline\n");

sol = ini->prox;

for (i = 0; i < n; i++)
{
  if (sol->v[i] == 0)
    fprintf (fp, "R ");

  for (j = 1; j < n; j++)
    {
      if (sol->v[i] == j)
	fprintf (fp, "& R ");
      else
	fprintf (fp, "& ");
    }
  fprintf (fp, "\\\\\n\\hline\n"); /* pula linha */
}

fprintf (fp, "\\end{tabular}\n");
fprintf (fp, "\\end{center}\n");

@ Para imprimirmos todas, percorremos a lista de solu\c c\~oes
e vamos imprimindo uma a uma (e j\'a aproveitamos e deletamos
para liberar o espa\c co).

@<Imprime t...@> =
fprintf (fp, "\\begin{itemize}\n");
for (k = 1; k <= num_sol; k++)
{
  fprintf (fp, "\\item Solu\\c c\\~ao %d:\n", k);
  @<Imprime u...@>@;
  free (sol->v);
  sol = sol->prox;
  free (ini->prox);
  ini->prox = sol;
}
fprintf (fp, "\\end{itemize}\n");

@ Fechando o arquivo.

@<Fecha...@> =
fprintf (fp, "\\end{document}\n");
fclose (fp);
if (todas)
{
  fprintf (stdout, "%d possibilidades foram escritas ", num_sol); 
  fprintf (stdout, "no arquivo %s.\n\n", nome_arq);
}
else
fprintf (stdout, "Resultado impresso no arquivo %s.\n\n", nome_arq);


@*1 Index. Aqui est\~ao todos os identificadores:
